@@ -1,19 +1,19 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os.path
 
-from xpra.log import Logger
-log = Logger("clipboard")
-
 from xpra.platform.features import CLIPBOARDS
-from xpra.util import csv
+from xpra.util import csv, nonl, XPRA_CLIPBOARD_NOTIFICATION_ID
 from xpra.scripts.config import FALSE_OPTIONS
 from xpra.server.mixins.stub_server_mixin import StubServerMixin
+from xpra.log import Logger
+
+log = Logger("clipboard")
 
 
 """
@@ -27,7 +27,7 @@ class ClipboardServer(StubServerMixin):
         self.clipboard_filter_file = None
 
     def init(self, opts):
-        self.clipboard = not ((opts.clipboard or "").lower() in FALSE_OPTIONS)
+        self.clipboard = not (opts.clipboard or "").lower() in FALSE_OPTIONS
         self.clipboard_direction = opts.clipboard_direction
         self.clipboard_filter_file = opts.clipboard_filter_file
 
@@ -47,6 +47,11 @@ class ClipboardServer(StubServerMixin):
             ch.client_reset()
 
 
+    def parse_hello(self, ss, _caps, send_ui):
+        if send_ui and self.clipboard:
+            self.parse_hello_ui_clipboard(ss)
+
+
     def get_info(self, _proto):
         if self._clipboard_helper is None:
             return {}
@@ -54,16 +59,18 @@ class ClipboardServer(StubServerMixin):
 
 
     def get_server_features(self, server_source=None):
-        clipboard = self._clipboard_helper is not None and self._clipboard_client == server_source
-        log("clipboard_helper=%s, clipboard_client=%s, source=%s, clipboard=%s", self._clipboard_helper, self._clipboard_client, server_source, clipboard)
+        clipboard = self._clipboard_helper is not None
+        log("clipboard_helper=%s, clipboard_client=%s, source=%s, clipboard=%s",
+            self._clipboard_helper, self._clipboard_client, server_source, clipboard)
         if not clipboard:
             return {}
         f = {
             "clipboards"            : self._clipboards,
             "clipboard-direction"   : self.clipboard_direction,
             "clipboard" : {
-                ""                  : True,
-                "enable-selections" : True,
+                ""                      : True,
+                "enable-selections"     : True,
+                "contents-slice-fix"    : True,
                 },
             }
         if self._clipboard_helper:
@@ -81,15 +88,18 @@ class ClipboardServer(StubServerMixin):
         clipboard_filter_res = []
         if self.clipboard_filter_file:
             if not os.path.exists(self.clipboard_filter_file):
-                log.error("invalid clipboard filter file: '%s' does not exist - clipboard disabled!", self.clipboard_filter_file)
+                log.error("invalid clipboard filter file: '%s' does not exist - clipboard disabled!",
+                          self.clipboard_filter_file)
                 return
             try:
                 with open(self.clipboard_filter_file, "r" ) as f:
                     for line in f:
                         clipboard_filter_res.append(line.strip())
-                    log("loaded %s regular expressions from clipboard filter file %s", len(clipboard_filter_res), self.clipboard_filter_file)
+                    log("loaded %s regular expressions from clipboard filter file %s",
+                        len(clipboard_filter_res), self.clipboard_filter_file)
             except:
-                log.error("error reading clipboard filter file %s - clipboard disabled!", self.clipboard_filter_file, exc_info=True)
+                log.error("Error: reading clipboard filter file %s - clipboard disabled!",
+                          self.clipboard_filter_file, exc_info=True)
                 return
         try:
             from xpra.clipboard.gdk_clipboard import GDKClipboardProtocolHelper
@@ -98,14 +108,16 @@ class ClipboardServer(StubServerMixin):
                       "can-send"    : self.clipboard_direction in ("to-client", "both"),
                       "can-receive" : self.clipboard_direction in ("to-server", "both"),
                       }
-            self._clipboard_helper = GDKClipboardProtocolHelper(self.send_clipboard_packet, self.clipboard_progress, **kwargs)
+            self._clipboard_helper = GDKClipboardProtocolHelper(self.send_clipboard_packet,
+                                                                self.clipboard_progress, **kwargs)
+            self._clipboard_helper.init_proxies_uuid()
             self._clipboards = CLIPBOARDS
         except Exception:
             #log("gdk clipboard helper failure", exc_info=True)
             log.error("Error: failed to setup clipboard helper", exc_info=True)
             self.clipboard = False
 
-    def parse_hello_ui_clipboard(self, ss, c):
+    def parse_hello_ui_clipboard(self, ss):
         #take the clipboard if no-one else has it yet:
         if not ss.clipboard_enabled:
             log("client does not support clipboard")
@@ -117,40 +129,62 @@ class ClipboardServer(StubServerMixin):
         if cc and not cc.is_closed():
             log("another client already owns the clipboard")
             return
+        self.set_clipboard_source(ss)
+
+    def set_clipboard_source(self, ss):
+        if self._clipboard_client==ss:
+            return
         self._clipboard_client = ss
-        self._clipboard_helper.init_proxies_uuid()
-        #deal with buggy win32 clipboards:
-        if "clipboard.greedy" not in c:
-            #old clients without the flag: take a guess based on platform:
-            client_platform = c.strget("platform", "")
-            greedy = client_platform.startswith("win") or client_platform.startswith("darwin")
+        ch = self._clipboard_helper
+        log("client %s is the clipboard peer, helper=%s", ss, ch)
+        if not ch:
+            return
+        if ss:
+            log(" greedy=%s", ss.clipboard_greedy)
+            log(" want targets=%s", ss.clipboard_want_targets)
+            log(" server has selections: %s", csv(self._clipboards))
+            log(" client initial selections: %s", csv(ss.clipboard_client_selections))
+            ch.set_greedy_client(ss.clipboard_greedy)
+            ch.set_want_targets_client(ss.clipboard_want_targets)
+            ch.enable_selections(ss.clipboard_client_selections)
+            ch.set_clipboard_contents_slice_fix(ss.clipboard_contents_slice_fix)
         else:
-            greedy = c.boolget("clipboard.greedy")
-        self._clipboard_helper.set_greedy_client(greedy)
-        want_targets = c.boolget("clipboard.want_targets")
-        self._clipboard_helper.set_want_targets_client(want_targets)
-        #the selections the client supports (default to all):
-        client_selections = c.strlistget("clipboard.selections", CLIPBOARDS)
-        log("client %s is the clipboard peer", ss)
-        log(" greedy=%s", greedy)
-        log(" want targets=%s", want_targets)
-        log(" server has selections: %s", csv(self._clipboards))
-        log(" client initial selections: %s", csv(client_selections))
-        self._clipboard_helper.enable_selections(client_selections)
+            ch.enable_selections([])
+
+
+    def last_client_exited(self):
+        ch = self._clipboard_helper
+        if ch:
+            ch.client_reset()
+
+
+    def set_session_driver(self, source):
+        self.set_clipboard_source(source)
+        ch = self._clipboard_helper
+        if not source or not ch:
+            return
+        log("set_session_driver(%s) clipboard_enabled=%s, clipboard helper=%s", source, source.clipboard_enabled, ch)
+        if source.clipboard_enabled:
+            log("selections: %s", source.clipboard_client_selections)
+            ch.send_tokens(source.clipboard_client_selections)
+
 
     def _process_clipboard_packet(self, proto, packet):
+        assert self.clipboard
         if self.readonly:
             return
         ss = self._server_sources.get(proto)
         if not ss:
             #protocol has been dropped!
             return
-        assert self._clipboard_client==ss, \
-                "the clipboard packet '%s' does not come from the clipboard owner!" % packet[0]
+        if self._clipboard_client!=ss:
+            log("the clipboard packet '%s' does not come from the clipboard owner!", packet[0])
+            return
         if not ss.clipboard_enabled:
             #this can happen when we disable clipboard in the middle of transfers
             #(especially when there is a clipboard loop)
-            log.warn("received a clipboard packet from a source which does not have clipboard enabled!")
+            log.warn("Warning: unexpected clipboard packet")
+            log.warn(" clipboard is disabled for %s", nonl(ss.uuid))
             return
         ch = self._clipboard_helper
         assert ch, "received a clipboard packet but clipboard sharing is disabled"
@@ -176,10 +210,15 @@ class ClipboardServer(StubServerMixin):
             #if we can, tell the client to do the same:
             if ss.clipboard_set_enabled:
                 ss.send_clipboard_enabled("probable clipboard loop detected")
+                body = "Too many clipboard requests,\n"+\
+                       "a clipboard synchronization loop is likely to be causing this problem"
+                ss.may_notify(XPRA_CLIPBOARD_NOTIFICATION_ID,
+                              "Clipboard synchronization is now disabled", body, icon_name="clipboard")
             return  False
         return True
 
     def _process_clipboard_enabled_status(self, proto, packet):
+        assert self.clipboard
         if self.readonly:
             return
         clipboard_enabled = packet[1]
@@ -203,12 +242,12 @@ class ClipboardServer(StubServerMixin):
         log("toggled clipboard to %s for %s", clipboard_enabled, ss.protocol)
 
     def clipboard_progress(self, local_requests, _remote_requests):
-        assert self._clipboard_helper is not None
+        assert self.clipboard
         if self._clipboard_client:
             self._clipboard_client.send_clipboard_progress(local_requests)
 
     def send_clipboard_packet(self, *parts):
-        assert self._clipboard_helper is not None
+        assert self.clipboard
         if self._clipboard_client:
             self._clipboard_client.send_clipboard(parts)
 
@@ -218,5 +257,8 @@ class ClipboardServer(StubServerMixin):
             self._authenticated_packet_handlers.update({
                 "set-clipboard-enabled":                self._process_clipboard_enabled_status,
               })
-            for x in ("token", "request", "contents", "contents-none", "pending-requests", "enable-selections", "loop-uuids"):
+            for x in (
+                "token", "request", "contents", "contents-none",
+                "pending-requests", "enable-selections", "loop-uuids",
+                ):
                 self._authenticated_ui_packet_handlers["clipboard-%s" % x] = self._process_clipboard_packet
