@@ -1,32 +1,29 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
-from collections import deque
 from math import sqrt
 from time import sleep
 
+from xpra.server.source.stub_source_mixin import StubSourceMixin
+from xpra.server.window.batch_config import DamageBatchConfig
+from xpra.server.server_core import ClientException
+from xpra.codecs.video_helper import getVideoHelper
+from xpra.codecs.codec_constants import video_spec
+from xpra.net.compression import compressed_wrapper, Compressed, use_lz4, use_lzo
+from xpra.os_util import monotonic_time, strtobytes
+from xpra.server.background_worker import add_work_item
+from xpra.util import csv, typedict, envint
 from xpra.log import Logger
+
 log = Logger("encoding")
 proxylog = Logger("proxy")
 statslog = Logger("stats")
 
-
-from xpra.server.source.stub_source_mixin import StubSourceMixin
-from xpra.server.window.batch_config import DamageBatchConfig
-from xpra.codecs.video_helper import getVideoHelper
-from xpra.codecs.codec_constants import video_spec
-from xpra.net.compression import compressed_wrapper, Compressed, use_lz4, use_lzo
-from xpra.make_thread import start_thread
-from xpra.os_util import Queue, monotonic_time
-from xpra.server.background_worker import add_work_item
-from xpra.util import csv, typedict, envint, envbool
-
-NOYIELD = not envbool("XPRA_YIELD", False)
 MIN_PIXEL_RECALCULATE = envint("XPRA_MIN_PIXEL_RECALCULATE", 2000)
 
 
@@ -36,23 +33,19 @@ Runs the encode thread.
 """
 class EncodingsMixin(StubSourceMixin):
 
-    def __init__(self, core_encodings, encodings, default_encoding, scaling_control,
-                 default_quality, default_min_quality,
-                 default_speed, default_min_speed):
-        log("ServerSource%s", (core_encodings, encodings, default_encoding, scaling_control,
-                 default_quality, default_min_quality,
-                 default_speed, default_min_speed))
-        self.server_core_encodings = core_encodings
-        self.server_encodings = encodings
-        self.default_encoding = default_encoding
-        self.scaling_control = scaling_control
+    def __init__(self):
+        self.server_core_encodings = []
+        self.server_encodings = []
+        self.default_encoding = None
+        self.scaling_control = None
 
-        self.default_quality = default_quality      #default encoding quality for lossy encodings
-        self.default_min_quality = default_min_quality #default minimum encoding quality
-        self.default_speed = default_speed          #encoding speed (only used by x264)
-        self.default_min_speed = default_min_speed  #default minimum encoding speed
+        self.default_quality = 40       #default encoding quality for lossy encodings
+        self.default_min_quality = 10   #default minimum encoding quality
+        self.default_speed = 40         #encoding speed (only used by x264)
+        self.default_min_speed = 10     #default minimum encoding speed
 
-        self.default_batch_config = DamageBatchConfig()     #contains default values, some of which may be supplied by the client
+        #contains default values, some of which may be supplied by the client:
+        self.default_batch_config = DamageBatchConfig()
         self.global_batch_config = self.default_batch_config.clone()      #global batch config
 
         self.vrefresh = -1
@@ -82,23 +75,23 @@ class EncodingsMixin(StubSourceMixin):
         #which may be used by other clients (other ServerSource instances)
         self.video_helper = getVideoHelper().clone()
 
-        # the queues of damage requests we work through:
-        self.encode_work_queue = Queue()            #holds functions to call to compress data (pixels, clipboard)
-                                                    #items placed in this queue are picked off by the "encode" thread,
-                                                    #the functions should add the packets they generate to the 'packet_queue'
-        self.packet_queue = deque()                 #holds actual packets ready for sending (already encoded)
-                                                    #these packets are picked off by the "protocol" via 'next_packet()'
-                                                    #format: packet, wid, pixels, start_send_cb, end_send_cb
-                                                    #(only packet is required - the rest can be 0/None for clipboard packets)
-        self.encode_thread = start_thread(self.encode_loop, "encode")
 
+    def init_from(self, _protocol, server):
+        self.server_core_encodings  = server.core_encodings
+        self.server_encodings       = server.encodings
+        self.default_encoding       = server.default_encoding
+        self.scaling_control        = server.scaling_control
+        self.default_quality        = server.default_quality
+        self.default_min_quality    = server.default_min_quality
+        self.default_speed          = server.default_speed
+        self.default_min_speed      = server.default_min_speed
 
     def cleanup(self):
         self.cancel_recalculate_timer()
         #Warning: this mixin must come AFTER the window mixin!
         #to make sure that it is safe to add the end of queue marker:
         #(all window sources will have stopped queuing data)
-        self.encode_work_queue.put(None)
+        self.queue_encode(None)
         #this should be a noop since we inherit an initialized helper:
         self.video_helper.cleanup()
 
@@ -147,9 +140,10 @@ class EncodingsMixin(StubSourceMixin):
         wids = tuple(self.calculate_window_ids)  #make a copy so we don't clobber new wids
         focus = self.get_focus()
         sources = self.window_sources.items()
-        maximized_wids = [wid for wid, source in sources if source is not None and source.maximized]
-        fullscreen_wids = [wid for wid, source in sources if source is not None and source.fullscreen]
-        log("recalculate_delays() wids=%s, focus=%s, maximized=%s, fullscreen=%s", wids, focus, maximized_wids, fullscreen_wids)
+        maximized_wids = tuple(wid for wid, source in sources if source is not None and source.maximized)
+        fullscreen_wids = tuple(wid for wid, source in sources if source is not None and source.fullscreen)
+        log("recalculate_delays() wids=%s, focus=%s, maximized=%s, fullscreen=%s",
+            wids, focus, maximized_wids, fullscreen_wids)
         for wid in wids:
             #this is safe because we only add to this set from other threads:
             self.calculate_window_ids.remove(wid)
@@ -182,17 +176,22 @@ class EncodingsMixin(StubSourceMixin):
             tsize += w*h
             tcount += 1
             time_w = 2.0+(now-ws.batch_config.last_updated)     #add 2 seconds to even things out
-            weight = w*h*time_w
+            weight = int(w*h*time_w)
             wdelay += ws.batch_config.delay*weight
             wdimsum += weight
         if wdimsum>0 and tcount>0:
             #weighted delay:
-            avg_size = tsize/tcount
-            wdelay = wdelay / wdimsum
-            #store the delay as a normalized value per megapixel:
-            delay = wdelay * 1000000 / avg_size
+            delay = wdelay // wdimsum
             self.global_batch_config.last_delays.append((now, delay))
             self.global_batch_config.delay = delay
+            #store the delay as a normalized value per megapixel
+            #so we can adapt it to different window sizes:
+            avg_size = tsize // tcount
+            ratio = sqrt(1000000.0 / avg_size)
+            normalized_delay = int(delay * ratio)
+            self.global_batch_config.delay_per_megapixel = normalized_delay
+            log("delay_per_megapixel=%i, delay=%i, for wdelay=%i, avg_size=%i, ratio=%.2f",
+                normalized_delay, delay, wdelay, avg_size, ratio)
 
     def may_recalculate(self, wid, pixel_count):
         if wid in self.calculate_window_ids:
@@ -211,7 +210,8 @@ class EncodingsMixin(StubSourceMixin):
         if delta>RECALCULATE_DELAY:
             add_work_item(self.recalculate_delays)
         else:
-            self.calculate_timer = self.timeout_add(int(1000*(RECALCULATE_DELAY-delta)), add_work_item, self.recalculate_delays)
+            delay = int(1000*(RECALCULATE_DELAY-delta))
+            self.calculate_timer = self.timeout_add(delay, add_work_item, self.recalculate_delays)
 
     def cancel_recalculate_timer(self):
         ct = self.calculate_timer
@@ -228,8 +228,8 @@ class EncodingsMixin(StubSourceMixin):
                 if value is not None:
                     try:
                         return int(value)
-                    except:
-                        log.error("invalid value for batch option %s: %s", varname, value)
+                    except (TypeError, ValueError):
+                        log.error("Error: invalid value '%s' for batch option %s", value, varname)
                 return None
             #from client caps first:
             cpname = "batch.%s" % prop
@@ -270,7 +270,7 @@ class EncodingsMixin(StubSourceMixin):
         self.encodings = c.strlistget("encodings")
         self.core_encodings = c.strlistget("encodings.core", self.encodings)
         if self.send_windows and not self.core_encodings:
-            raise Exception("client failed to specify any supported encodings")
+            raise ClientException("client failed to specify any supported encodings")
         if "png" in self.core_encodings:
             self.window_icon_encodings.append("png")
         self.window_icon_encodings = c.strlistget("encodings.window-icon", ["premult_argb32"])
@@ -299,8 +299,10 @@ class EncodingsMixin(StubSourceMixin):
                 self.encoding_options[ek] = c.boolget(k)
         #2: standardized encoding options:
         for k in c.keys():
-            if k.startswith(b"theme.") or  k.startswith(b"encoding.icons."):
-                self.icons_encoding_options[k.replace(b"encoding.icons.", b"").replace(b"theme.", b"")] = c[k]
+            #yaml gives us str..
+            k = strtobytes(k)
+            if k.startswith(b"theme.") or k.startswith(b"encoding.icons."):
+                self.icons_encoding_options[k.replace(b"encoding.icons.", b"").replace(b"theme.", b"")] = c.get(k)
             elif k.startswith(b"encoding."):
                 stripped_k = k[len(b"encoding."):]
                 if stripped_k in (b"transparency",
@@ -327,7 +329,11 @@ class EncodingsMixin(StubSourceMixin):
             except:
                 proxylog.error("failed to parse proxy video", exc_info=True)
 
-        self.default_encoding_options["scaling.control"] = self.encoding_options.get("scaling.control", self.scaling_control)
+        sc = self.encoding_options.get("scaling.control", self.scaling_control)
+        if sc is not None:
+            #"encoding_options" are exposed via "xpra info",
+            #so we can't have None values in there (bencoder would choke)
+            self.default_encoding_options["scaling.control"] = sc
         q = self.encoding_options.intget("quality", self.default_quality)         #0.7 onwards:
         if q>0:
             self.default_encoding_options["quality"] = q
@@ -342,8 +348,10 @@ class EncodingsMixin(StubSourceMixin):
             self.default_encoding_options["min-speed"] = ms
         log("default encoding options: %s", self.default_encoding_options)
         self.auto_refresh_delay = c.intget("auto_refresh_delay", 0)
-        if self.mmap_size==0:
-            others = [x for x in self.core_encodings if x in self.server_core_encodings and x!=self.encoding]
+        #check for mmap:
+        if getattr(self, "mmap_size", 0)==0:
+            others = tuple(x for x in self.core_encodings
+                           if x in self.server_core_encodings and x!=self.encoding)
             if self.encoding=="auto":
                 s = "automatic picture encoding enabled"
             else:
@@ -363,8 +371,23 @@ class EncodingsMixin(StubSourceMixin):
             for colorspace, spec_props in colorspace_specs.items():
                 for spec_prop in spec_props:
                     #make a new spec based on spec_props:
-                    spec = video_spec(codec_class=Encoder, codec_type="proxy", encoding=encoding)
+                    spec_prop = typedict(spec_prop)
+                    input_colorspace = spec_prop.strget("input_colorspace")
+                    output_colorspaces = spec_prop.strlistget("output_colorspaces")
+                    if not input_colorspace or not output_colorspaces:
+                        log.warn("Warning: invalid proxy video encoding '%s':", encoding)
+                        log.warn(" missing colorspace attributes")
+                        continue
+                    spec = video_spec(codec_class=Encoder,
+                                      has_lossless_mode=spec_prop.boolget("has_lossless_mode", False),
+                                      input_colorspace=input_colorspace,
+                                      output_colorspaces=output_colorspaces,
+                                      codec_type="proxy", encoding=encoding,
+                                      )
                     for k,v in spec_prop.items():
+                        if k.startswith("_") or not hasattr(spec, k):
+                            log.warn("Warning: invalid proxy codec attribute '%s'", k)
+                            continue
                         setattr(spec, k, v)
                     proxylog("parse_proxy_video() adding: %s / %s / %s", encoding, colorspace, spec)
                     self.video_helper.add_encoder_spec(encoding, colorspace, spec)
@@ -428,7 +451,7 @@ class EncodingsMixin(StubSourceMixin):
         ieo = dict(self.icons_encoding_options)
         try:
             del ieo["default.icons"]
-        except:
+        except KeyError:
             pass
         #encoding:
         info.update({
@@ -439,9 +462,11 @@ class EncodingsMixin(StubSourceMixin):
                                            },
                      "icons"            : ieo,
                      })
-        einfo = {"default"      : self.default_encoding or ""}
-        einfo.update(self.default_encoding_options)
-        einfo.update(self.encoding_options)
+        einfo = {
+            "default"      : self.default_encoding or "",
+            "defaults"     : self.default_encoding_options,
+            "client-defaults" : self.encoding_options,
+            }
         info.setdefault("encoding", {}).update(einfo)
         return info
 
@@ -484,63 +509,11 @@ class EncodingsMixin(StubSourceMixin):
         #but use sqrt to smooth things and prevent excesses
         #(ie: a 4MPixel window, will start at 2 times the global delay)
         #(ie: a 0.5MPixel window will start at 0.7 times the global delay)
+        dpm = self.global_batch_config.delay_per_megapixel
         w, h = window.get_dimensions()
-        ratio = float(w*h) / 1000000
-        batch_config.delay = self.global_batch_config.delay * sqrt(ratio)
+        if dpm>=0:
+            ratio = sqrt(1000000.0 / (w*h))
+            batch_config.delay = max(batch_config.min_delay, min(batch_config.max_delay, int(dpm * sqrt(ratio))))
+        log("make_batch_config(%i, %s) global delay per megapixel=%i, new window delay for %ix%i=%s",
+                 wid, window, dpm, w, h, batch_config.delay)
         return batch_config
-
-
-    def queue_size(self):
-        return self.encode_work_queue.qsize()
-
-    def call_in_encode_thread(self, *fn_and_args):
-        """
-            This is used by WindowSource to queue damage processing to be done in the 'encode' thread.
-            The 'encode_and_send_cb' will then add the resulting packet to the 'packet_queue' via 'queue_packet'.
-        """
-        self.statistics.compression_work_qsizes.append((monotonic_time(), self.encode_work_queue.qsize()))
-        self.encode_work_queue.put(fn_and_args)
-
-    def queue_packet(self, packet, wid=0, pixels=0, start_send_cb=None, end_send_cb=None, fail_cb=None, wait_for_more=False):
-        """
-            Add a new 'draw' packet to the 'packet_queue'.
-            Note: this code runs in the non-ui thread
-        """
-        now = monotonic_time()
-        self.statistics.packet_qsizes.append((now, len(self.packet_queue)))
-        if wid>0:
-            self.statistics.damage_packet_qpixels.append((now, wid, sum(x[2] for x in tuple(self.packet_queue) if x[1]==wid)))
-        self.packet_queue.append((packet, wid, pixels, start_send_cb, end_send_cb, fail_cb, wait_for_more))
-        p = self.protocol
-        if p:
-            p.source_has_more()
-
-#
-# The damage packet thread loop:
-#
-    def encode_loop(self):
-        """
-            This runs in a separate thread and calls all the function callbacks
-            which are added to the 'encode_work_queue'.
-            Must run until we hit the end of queue marker,
-            to ensure all the queued items get called.
-        """
-        while True:
-            fn_and_args = self.encode_work_queue.get(True)
-            if fn_and_args is None:
-                return              #empty marker
-            #some function calls are optional and can be skipped when closing:
-            #(but some are not, like encoder clean functions)
-            optional_when_closing = fn_and_args[0]
-            if optional_when_closing and self.is_closed():
-                continue
-            try:
-                fn_and_args[1](*fn_and_args[2:])
-            except Exception as e:
-                if self.is_closed():
-                    log("ignoring encoding error in %s as source is already closed:", fn_and_args[0])
-                    log(" %s", e)
-                else:
-                    log.error("Error during encoding:", exc_info=True)
-                del e
-            NOYIELD or sleep(0)
